#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from iambic import ast


def render_json(tree: ast.Play) -> str:
    """The renderer for a JSON document.

    This callable accepts a :class:`ast.PlayTree` and returns a valid JSON document.
    It's mainly included for api-completion, since it actually just accesses already public properties.

    Parameters
    ----------
    tree
        The play tree, as loaded via :func:`iambic.parse.text` or :func:`iambic.parse.data`
    """
    return tree.json()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import enum
from typing import Union, Mapping

import inflection

from iambic import ast

import yattag


class HTMLTag(str, enum.Enum):
    A = "a"
    BODY = "body"
    BR = "br"
    DIV = "div"
    DTYPE = "<!DOCTYPE html>"
    EM = "em"
    H1 = "h1"
    H2 = "h2"
    H3 = "h3"
    H4 = "h4"
    H5 = "h5"
    HDR = "header"
    HEAD = "head"
    HREF = "href"
    HTML = "html"
    LI = "li"
    META = "meta"
    MTR = "meter"
    NAV = "nav"
    OL = "ol"
    P = "p"
    SECT = "section"
    SPAN = "span"
    STR = "strong"
    TITLE = "title"


class TOC:
    tag = HTMLTag.H2
    title = "Table of Contents"
    klass = inflection.parameterize(title)


class HTMLRenderer:
    """The Renderer for an HTML document.

    This is a callable class which accepts a :class:`ast.PlayTree` and generates an HTML document with attached
    meta-data for easy querying and CSS styling (not provided by default).
    """

    tag = HTMLTag

    def render_speech(
        self,
        speech: ast.Speech,
        personae: Mapping[str, ast.Persona],
        doc: yattag.SimpleDoc,
    ):
        """Render a "speech" node.

        A :class:`ast.SpeechNode` is a container for all lines of dialogue, character action, or stage direction which
        can be attributed to an individual section of speech for a character.

        Parameters
        ----------
        speech
            The speech node to render to html.
        doc
            The document object which the spech will be rendered to.
        personae
            A mapping of :py:attr:`Persona.id` -> :py:class:`Persona`
        """
        persona = personae[speech.persona]
        doc.line(
            self.tag.STR.value,
            persona.name,
            klass=persona.klass,
            id=f"{persona.id}-{speech.id}",
        )
        doc.stag(self.tag.BR.value)
        last_line = None
        for line in speech.speech:
            with doc.tag(self.tag.SPAN.value, klass=line.klass, id=line.id):
                doc.data(index=line.index)
                if isinstance(line, ast.Action):
                    if last_line:
                        doc.stag(self.tag.BR.value)
                    doc.line(self.tag.EM.value, f"[{line.action}]")
                elif isinstance(line, ast.Direction):
                    doc.stag(self.tag.BR.value)
                    doc.line(self.tag.EM.value, line.action)
                    doc.stag(self.tag.BR.value)
                else:
                    doc.data(lineno=line.lineno)
                    doc.text(line.line)
                doc.stag(self.tag.BR.value)
            last_line = line

    def render_scene(
        self,
        scene: ast.NodeTree,
        doc: yattag.SimpleDoc,
        personae: Mapping[str, ast.Persona],
        *,
        as_act: bool = False,
    ):
        """Render a "scene". This could be either a :class:`ast.NodeTree` or :class:`ast.Intermission`

        Parameters
        ----------
        scene
            A child of the higher-level "act" -
            which could be a :class:`ast.Act`, :class:`ast.Epilogue`, or :class:`ast.Prologue`
        doc
            The initialized document object.
        personae
            A mapping of :py:attr:`Persona.id` -> :py:class:`Persona`
        as_act : default False
            Optionally render this scene at the "act" level (:class:`HTMLTag.H1`)
        """
        if isinstance(scene, ast.Intermission):
            with doc.tag(self.tag.H1.value, klass=scene.klass, id=scene.id):
                doc.text(scene.text)
                doc.data(index=scene.index)
        else:
            text = scene.node.text
            node: ast.Scene = scene.node
            tag = self.tag.H1.value if as_act else self.tag.H3.value
            with doc.tag(tag, klass=node.klass, id=node.id):
                doc.text(text)
                doc.data(index=node.index)
            for line in scene.children:
                with doc.tag(self.tag.P.value, klass=line.klass, id=line.id):
                    if type(line) in {ast.Direction, ast.Entrance, ast.Exit}:
                        with doc.tag(self.tag.EM.value):
                            doc.text(
                                line.action
                                if isinstance(line, ast.Direction)
                                else line.text
                            )
                            doc.data(index=line.index)
                        continue
                    self.render_speech(line, personae, doc)

    def render_play(self, tree: ast.Play) -> yattag.SimpleDoc:
        doc = yattag.SimpleDoc()
        persona_map = {x.id: x for x in tree.personae}
        for act in tree.children:
            if isinstance(act.node, (ast.Epilogue, ast.Prologue)):
                self.render_scene(act, doc, persona_map, as_act=True)
                continue
            with doc.tag(self.tag.H1.value, klass=act.node.klass, id=act.node.id):
                doc.data(index=act.node.index)
                doc.text(act.node.text)
            for scene in act.children:
                self.render_scene(scene, doc, persona_map)
        return doc

    def render_toc(self, tree: ast.Play) -> yattag.SimpleDoc:
        toc = yattag.SimpleDoc()
        with toc.tag(
            self.tag.NAV.value,
            klass=TOC.klass,
            id=f"{inflection.parameterize(str(tree.meta.title))}-{TOC.klass}",
            role="navigation",
        ):
            toc.line(TOC.tag.value, TOC.title)
            with toc.tag(self.tag.OL.value):
                for act in tree.children:
                    with toc.tag(self.tag.LI.value):
                        toc.line(
                            self.tag.A.value, act.node.text, href=f"#{act.node.id}"
                        )
                        if isinstance(act.node, ast.Act):
                            with toc.tag(self.tag.OL.value):
                                for scene in act.children:
                                    with toc.tag(self.tag.LI.value):
                                        node = getattr(scene, "node", scene)
                                        toc.line(
                                            self.tag.A.value,
                                            node.text,
                                            href=f"#{node.id}",
                                        )
        return toc

    def render_head(self, tree: ast.Play) -> yattag.SimpleDoc:
        head = yattag.SimpleDoc()
        with head.tag(self.tag.HEAD.value):
            head.line(self.tag.TITLE.value, tree.meta.title or "")
            head.stag(self.tag.META.value, name="language", content=tree.meta.language)
            head.stag(
                self.tag.META.value, name="publisher", content=tree.meta.publisher
            )
            head.stag(self.tag.META.value, name="rights", content=tree.meta.rights)
            head.stag(self.tag.META.value, name="author", content=tree.meta.author)
            head.stag(self.tag.META.value, name="edition", content=tree.meta.edition)
            head.stag(
                self.tag.META.value, name="editors", content=",".join(tree.meta.editors)
            )
            head.stag(
                self.tag.META.value, name="keywords", content=",".join(tree.meta.tags)
            )
            head.stag(
                self.tag.META.value,
                name="viewport",
                content="width=device-width, initial-scale=1.0",
            )

        return head

    def render(
        self, tree: ast.Play, as_text: bool = True, with_toc: bool = True
    ) -> Union[yattag.SimpleDoc, str]:
        """Render a :class:`ast.PlayTree` into an HTML document.

        Parameters
        ----------
        tree :
            The play tree, as loaded via :func:`~iambic.parser.parse` or :func:`~iambic.loader.load`
        as_text : default True
            Optionally return the raw :class:`yattag.SimpleDoc` if set to False.
        with_toc : default True
            Optionally turn off TOC rendering.
        """
        head = self.render_head(tree)
        play = self.render_play(tree)
        toc = self.render_toc(tree) if with_toc else yattag.SimpleDoc()
        doc = yattag.SimpleDoc()
        doc.asis(self.tag.DTYPE)
        with doc.tag(self.tag.HTML.value, lang="en"):
            doc.result.extend(head.result)
            with doc.tag(self.tag.BODY.value):
                doc.result.extend(toc.result)
                doc.result.extend(play.result)

        return yattag.indent(doc.getvalue()) if as_text else doc

    # Assign the highest-level method to the public API
    __call__ = render


# Pin the callable as the public API
render_html = HTMLRenderer()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import enum
import functools
import inspect
import ujson as json
import re
from typing import Mapping, Collection

from iambic import schema


__all__ = (
    "NodeType",
    "NodeToken",
    "NodeMixin",
    "DEFINITIONS",
    "jsonify",
    "NODE_PATTERN",
)


class NodeType(str, enum.Enum):
    """An enumeration of the different types of nodes in a script."""

    ACT = "act"
    SCENE = "scene"
    PROL = "prologue"
    EPIL = "epilogue"
    INTER = "intermission"
    PERS = "persona"
    ENTER = "entrance"
    EXIT = "exit"
    ACTION = "action"
    DIR = "direction"
    DIAL = "dialogue"
    SPCH = "speech"
    TREE = "tree"
    PLAY = "play"
    META = "meta"


class NodeToken(str, enum.Enum):
    """An enumeration of special 'tokens', e.g., strings of characters."""

    JOIN1 = "…"
    JOIN2 = "..."


NODE_PATTERN = re.compile(
    r"""
    (
        # Locales: Act, Scene, Prologue, Epilogue, Intermission
        ^\#+\s(?P<act>(ACT)\s([IVX]+|\d+))              |
        ^\#+\s(?P<scene>SCENE\s([IVX]+|\d+)).*          |
        ^\#+\s(?P<prologue>PROLOGUE).*                  |
        ^\#+\s(?P<epilogue>EPILOGUE).*                  |
        ^\#+\s(?P<intermission>(INTERMISSION)).*        |
        # Persona
        [*_]{2}(?P<persona>(
            ([A-Z][a-zA-Z'’]*\W{0,2}([a-z]+)?\s?
        )+([A-Z]|\d)*))[*_]{2}                          |
        # Enter/Exit/Action/Direction
        (
            (?P<start>^[_*]\[?)?
                (
                    (
                        (?P<entrance>(enter)((?![_*\[\]]).)*)|(?P<exit>(exeunt|exit)((?![_*\[\]]).)*)
                    ) 
                    |
                    (?P<direction>((?![_*\[\]]).)+)
                )
                
            (?P<end>\]?[_*])?\s{0,2}
        )                                               |
        # Dialogue (catch-all)
        (?P<dialogue>(^.+))
    )
    """,
    re.VERBOSE | re.IGNORECASE | re.MULTILINE,
)


def asdict(obj):
    if isinstance(obj, (str, bytes, int)):
        return obj

    result = {}
    for attr in (
        x
        for x in dir(obj)
        if not x.startswith("_")
        and x not in {"id", "klass", "linerange", "col", "cols", "num_lines"}
    ):
        val = getattr(obj, attr)
        if inspect.ismethod(val):
            continue
        if hasattr(val, "asdict"):
            val = val.asdict()
        elif isinstance(val, enum.Enum):
            val = val.value
        elif isinstance(val, Mapping):
            val = {x: asdict(y) for x, y in val.items()}
        elif isinstance(val, Collection) and not isinstance(val, str):
            val = tuple(asdict(x) for x in val)

        result[attr] = val

    return result


class NodeMixin:
    @property
    @functools.lru_cache(1)
    def klass(self):
        return type(self).__name__.lower()

    asdict = asdict


DEFINITIONS = schema.SCHEMA["definitions"]
jsonify = functools.partial(json.dumps, indent=4, separators=(", ", ": "))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from iambic import parse, render
from tests.static import MD_RAW


def test_json():
    parsed = parse.text(MD_RAW)
    assert render.json(parsed) == parsed.json()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import json
import pathlib

from .frozendict import frozendict
from .dataschema import dataschema

SCHEMA_PATH = pathlib.Path(__file__).resolve().parent / "schema.json"
SCHEMA = json.loads(SCHEMA_PATH.read_text())



from iambic.ast import *

PARSED = Play(
    children=(
        NodeTree(
            node=Prologue(index=0, text="Prologue", setting=None, act=None),
            children=(
                Entrance(index=1, text="Enter FOO", scene="prologue", personae=()),
                Speech(
                    persona="foo",
                    scene="prologue",
                    speech=(
                        Dialogue(
                            line="Bar.",
                            persona="foo",
                            scene="prologue",
                            index=3,
                            lineno=1,
                            linepart=0,
                        ),
                    ),
                    index=3,
                ),
                Speech(
                    persona="a-bar",
                    scene="prologue",
                    speech=(
                        Dialogue(
                            line="Bar!",
                            persona="a-bar",
                            scene="prologue",
                            index=5,
                            lineno=2,
                            linepart=0,
                        ),
                    ),
                    index=5,
                ),
                Speech(
                    persona="bar-s-foo",
                    scene="prologue",
                    speech=(
                        Dialogue(
                            line="Foo!",
                            persona="bar-s-foo",
                            scene="prologue",
                            index=7,
                            lineno=3,
                            linepart=0,
                        ),
                    ),
                    index=7,
                ),
                Exit(index=8, text="Exeunt", scene="prologue", personae=()),
            ),
            personae=("foo", "bar-s-foo", "a-bar"),
        ),
        NodeTree(
            node=Act(index=9, text="ACT I", num=1),
            children=(
                NodeTree(
                    node=Prologue(index=10, text="Prologue", setting=None, act="act-i"),
                    children=(
                        Entrance(
                            index=11,
                            text="Enter BAR",
                            scene="act-i-prologue",
                            personae=(),
                        ),
                        Speech(
                            persona="bar",
                            scene="act-i-prologue",
                            speech=(
                                Action(
                                    action="Aside",
                                    persona="bar",
                                    scene="act-i-prologue",
                                    index=13,
                                ),
                                Dialogue(
                                    line="Foo.",
                                    persona="bar",
                                    scene="act-i-prologue",
                                    index=14,
                                    lineno=4,
                                    linepart=0,
                                ),
                            ),
                            index=13,
                        ),
                    ),
                    personae=("bar",),
                ),
                Intermission(index=15, text="INTERMISSION", act="act-i"),
                NodeTree(
                    node=Scene(
                        index=16, text="SCENE I", num=1, act="act-i", setting=None
                    ),
                    children=(
                        Entrance(
                            index=17,
                            text="Enter FOO and BAR",
                            scene="act-i-scene-i",
                            personae=(),
                        ),
                        Speech(
                            persona="foo",
                            scene="act-i-scene-i",
                            speech=(
                                Action(
                                    action="To BAR",
                                    persona="foo",
                                    scene="act-i-scene-i",
                                    index=19,
                                ),
                                Dialogue(
                                    line="Bar.",
                                    persona="foo",
                                    scene="act-i-scene-i",
                                    index=20,
                                    lineno=5,
                                    linepart=0,
                                ),
                            ),
                            index=19,
                        ),
                        Speech(
                            persona="bar",
                            scene="act-i-scene-i",
                            speech=(
                                Dialogue(
                                    line="Foo.",
                                    persona="bar",
                                    scene="act-i-scene-i",
                                    index=22,
                                    lineno=6,
                                    linepart=0,
                                ),
                            ),
                            index=22,
                        ),
                        Speech(
                            persona="foo",
                            scene="act-i-scene-i",
                            speech=(
                                Dialogue(
                                    line="Bar. ...",
                                    persona="foo",
                                    scene="act-i-scene-i",
                                    index=24,
                                    lineno=7,
                                    linepart=0,
                                ),
                            ),
                            index=24,
                        ),
                        Speech(
                            persona="bar",
                            scene="act-i-scene-i",
                            speech=(
                                Dialogue(
                                    line="... Foo. ...",
                                    persona="bar",
                                    scene="act-i-scene-i",
                                    index=26,
                                    lineno=7,
                                    linepart=0,
                                ),
                            ),
                            index=26,
                        ),
                        Speech(
                            persona="foo",
                            scene="act-i-scene-i",
                            speech=(
                                Dialogue(
                                    line="... Bar.",
                                    persona="foo",
                                    scene="act-i-scene-i",
                                    index=28,
                                    lineno=7,
                                    linepart=0,
                                ),
                                Direction(
                                    action="A long stage direction that goes on forever and says far to much to be reasonable. So much so that it takes up more than two lines, which is just ridiculous. I mean seriously, why? Why write so much for one stage direction? For what purpose?",
                                    scene="act-i-scene-i",
                                    index=29,
                                    stop=True,
                                ),
                            ),
                            index=28,
                        ),
                    ),
                    personae=("foo", "bar"),
                ),
                NodeTree(
                    node=Epilogue(index=33, text="EPILOGUE", setting=None, act="act-i"),
                    children=(
                        Entrance(
                            index=34,
                            text="Enter FOO",
                            scene="act-i-epilogue",
                            personae=(),
                        ),
                        Speech(
                            persona="foo",
                            scene="act-i-epilogue",
                            speech=(
                                Action(
                                    action="Reads",
                                    persona="foo",
                                    scene="act-i-epilogue",
                                    index=36,
                                ),
                                Dialogue(
                                    line="Foo.",
                                    persona="foo",
                                    scene="act-i-epilogue",
                                    index=37,
                                    lineno=8,
                                    linepart=0,
                                ),
                            ),
                            index=36,
                        ),
                    ),
                    personae=("foo",),
                ),
            ),
            personae=(),
        ),
    ),
    personae=(
        Persona(index=2, text="FOO", name="Foo", short=None),
        Persona(index=4, text="A Bar", name="A Bar", short=None),
        Persona(index=6, text="Bar's Foo", name="Bar's Foo", short=None),
        Persona(index=12, text="BAR", name="Bar", short=None),
    ),
    meta=MetaData(),
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import enum

import typic

from .base import *  # noqa: F403 (we've defined __all__)
from .node import *  # noqa: F403 (we've defined __all__)
from .index import *  # noqa: F403 (we've defined __all__)


class InputType(str, enum.Enum):
    HTML = "html"
    MD = "markdown"
    DATA = "data"


typic.register(coercer=node_coercer, check=isnodetype, check_origin=False)
typic.resolve()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from .html import render_html as _render_html
from .json import render_json as _render_json
from .markdown import render_markdown as _render_markdown
from .table import tabulate as _tabulate

markdown = _render_markdown
html = _render_html
json = _render_json
table = _tabulate



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import pathlib

from .table import TABLE, MATRIX
from .parsed import PARSED

DIR = pathlib.Path(__file__).parent
HTML = (DIR / "foo.html").read_text()
MD_RAW = (DIR / "foo-raw.md").read_text()
JSON = (DIR / "foo.json").read_text()
DATASET = (DIR / "dataset").read_text().rstrip("\n")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools
import ujson as json

from iambic import ast


class Parser:
    __resolver_map__ = ast.GenericNode.__resolver_map__

    @functools.lru_cache()
    def parse(self, data: str) -> ast.ResolvedNode:
        data = json.loads(data) if isinstance(data, str) else data
        handler = self.__resolver_map__[data.pop("type")]
        return handler(**data)

    __call__ = parse


parser = Parser()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools
import pathlib
from typing import Set, Iterator, Optional, Union

import inflection

from iambic import parse, ast


class Corpus:

    PATH: pathlib.Path = pathlib.Path(__file__).parent

    @staticmethod
    def _get_names(path: pathlib.Path) -> Set[str]:
        names = {path.name}
        for child in path.iterdir():
            if child.suffix == ".md":
                names.add(inflection.parameterize(child.stem))
        return names

    def __iter__(self) -> Iterator[pathlib.Path]:
        yield from self.PATH.iterdir()

    @functools.lru_cache(maxsize=128)
    def __getitem__(self, item: str):
        if not isinstance(item, str):
            raise TypeError("Corpus keys must be type str, provided")
        seen = set()
        item = inflection.parameterize(item)
        for child in self:
            if child.is_dir():
                seen.add(child.name)
                names = self._get_names(child)
                if item in names or any(item in x for x in names):
                    for text in child.iterdir():
                        if text.suffix.endswith(".md"):
                            return text.read_text()
        raise KeyError(
            f"Couldn't locate a play with key <{item}>. Available keys are: {tuple(seen)}"
        )

    @functools.lru_cache(maxsize=128)
    def get(
        self, item: str, default: str = None, *, parsed: bool = False, tree: bool = True
    ) -> Optional[Union[str, ast.Play, ast.Index]]:
        try:
            text = self[item]
            return parse.text(text, tree=tree) if parsed else text
        except KeyError:
            return default


corpus = Corpus()
get = corpus.get



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import dataclasses
import functools
import logging
import typing
from collections import defaultdict
from typing import Pattern, Match

from html2text import html2text

from iambic.ast import (
    GenericNode,
    NodeType,
    Play,
    InputType,
    ResolvedNode,
    Index,
    NodeToken,
    NODE_PATTERN,
)
from iambic.schema import frozendict

logger = logging.getLogger(__name__)


@functools.lru_cache(maxsize=None, typed=True)
def _cached_match(pattern: Pattern, value: str) -> Match:
    return pattern.fullmatch(value) or pattern.match(value)


def _safe_resolve(
    node: typing.Optional[GenericNode], *, attr: str = None
) -> ResolvedNode:
    if node:
        resolved = node.resolve()
        return getattr(resolved, attr, None) if attr else resolved


_safe_id = functools.partial(_safe_resolve, attr="id")


@dataclasses.dataclass
class _PreNode:
    type: NodeType
    match: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    text: str = None
    index: int = 0
    lineno: int = 0
    linepart: int = 0

    def to_generic(self, **kwargs) -> GenericNode:
        return GenericNode(
            type=self.type,
            text=self.text,
            index=self.index,
            lineno=self.lineno,
            linepart=self.linepart,
            match=frozendict(self.match),
            **kwargs,
        )


@dataclasses.dataclass
class ParserContext:
    parent: GenericNode = None
    act: GenericNode = None
    scene: GenericNode = None
    character: GenericNode = None
    index: Index = dataclasses.field(default_factory=Index)
    lineno: int = 0
    linepart: int = 0


class Parser:
    """The text parser for generating a query-able data structure from a body of text.

    Text should be in valid markdown and uses the regex patterns defined in :py:class:`NodePattern`.
    """

    LOCALES = frozenset((NodeType.ACT, NodeType.PROL, NodeType.EPIL, NodeType.SCENE))
    _ACTION_BOOKENDS = frozenset(("[", "]"))

    def __init__(self):
        self.__parser_map = defaultdict(lambda: self.default_handler)
        # Map the custom handlers for different Node sub-types
        for typ in self.LOCALES:
            self.__parser_map[typ] = self.locale_handler
        self.__parser_map[NodeType.PERS] = self.persona_handler

    @classmethod
    @functools.lru_cache(maxsize=None, typed=True)
    def match(cls, line: str, index: int) -> _PreNode:
        """Take an individual line from a body of text and determine which :py:class:`NodeType` it is."""
        match = NODE_PATTERN.match(line)
        match_dict = (
            {x: y for x, y in match.groupdict().items() if y is not None}
            if match
            else {}
        )
        node = _PreNode(
            type=NodeType.DIAL, match={NodeType.DIAL: line}, text=line, index=index
        )
        for node_type in {*NodeType} & match_dict.keys():
            node_type = NodeType(node_type)
            if node_type == NodeType.DIR:
                bookend = match_dict.get("start") or match_dict.get("end")
                if bookend:
                    if cls._ACTION_BOOKENDS & {*bookend}:
                        node_type = NodeType.ACTION
                        match_dict[NodeType.ACTION] = match_dict.pop(NodeType.DIR)
                    node.match, node.type = match_dict, node_type
                    break
            else:
                node.match, node.type = match_dict, node_type
                break

        return node

    @staticmethod
    def default_handler(ctx: ParserContext, node: _PreNode) -> ParserContext:
        """The default node handler. Assumes a lower-level node.

        Takes a pre-node object and resolves it into a :py:class:`GenericNode`.
        Does some juggling of values in ``ctx`` to determine the correct values
        to provide to :py:class:`GenericNode` on init.
        """
        # determine the 'real parent' for this node
        # Set the current parent as the parent for this particular node.
        kwargs = {"parent": _safe_id(ctx.parent)}
        # We associate dialogue and character actions to the character & scene.
        if node.type in {NodeType.DIAL, NodeType.ACTION}:
            kwargs["parent"] = _safe_id(ctx.character)
            kwargs["scene"] = _safe_id(ctx.scene)
        # Intermissions can be associated to the act they occur within
        elif node.type == NodeType.INTER:
            kwargs["parent"] = _safe_id(ctx.act)

        node = node.to_generic(**kwargs)
        ctx.index.append(node)
        return ctx

    @staticmethod
    def persona_handler(ctx: ParserContext, node: _PreNode) -> ParserContext:
        """The handler for :py:class:`NodeType.PERS`"""
        node = node.to_generic()
        resolved = ctx.index[node]
        if resolved:
            node = dataclasses.replace(node, index=resolved.index)
        ctx.character = node
        ctx.index.append(node)
        return ctx

    @staticmethod
    def check_parent(
        act: typing.Optional[GenericNode], node: _PreNode
    ) -> typing.Tuple[GenericNode, GenericNode]:
        """Check if the current node qualifies as a 'Parent Node'."""
        kwargs = {}
        if node.type != NodeType.ACT:
            kwargs["parent"] = _safe_id(act)
        node = node.to_generic(**kwargs)
        parent = node
        return parent, node

    @staticmethod
    def check_linecount(ctx: ParserContext, node: _PreNode):
        """Check the current line-count and line-part of a given node & context."""
        if node.type == NodeType.DIAL:
            text = node.text.strip()
            continued = text.endswith(NodeToken.JOIN1) or text.endswith(NodeToken.JOIN2)
            continuing = text.startswith(NodeToken.JOIN1) or text.startswith(
                NodeToken.JOIN2
            )
            # started a line-continuation or isn't a line-continuation
            if (continued and not continuing) or (not continued and not continuing):
                node.lineno = ctx.lineno = ctx.lineno + 1
            # in a line-continuation
            if continued or continuing:
                node.linepart = ctx.linepart = ctx.linepart + 1
                node.lineno = ctx.lineno
            # finished a line-continuation
            if not continued:
                ctx.linepart = 0

    @classmethod
    def check_direction(cls, ctx: ParserContext, node: _PreNode) -> bool:
        """Check that a given node is not actually within a stage-direction/action/enter/exit context."""
        append = True
        if ctx.index:
            prev: GenericNode = ctx.index.generic[-1]

            if (
                prev.type
                in {NodeType.DIR, NodeType.ACTION, NodeType.EXIT, NodeType.ENTER}
                and node.type in {NodeType.DIAL, prev.type}
                and not prev.match.get("end")
            ):
                text = f"{prev.text.strip()} {node.text.strip()}"
                node = cls.match(text, prev.index)
                ctx.index[-1] = dataclasses.replace(
                    prev, match=frozendict(node.match), text=text
                )
                append = False
        return append

    def locale_handler(self, ctx: ParserContext, node: _PreNode) -> ParserContext:
        """The handler for locale-types, defined in :py:attr:`Parser.LOCALES`"""
        ctx.parent, node = self.check_parent(ctx.act, node)
        last_type = ctx.index[-1].type if ctx.index else None
        if node.type == NodeType.ACT:
            ctx.act = node
        elif node.type in {NodeType.EPIL, NodeType.PROL} and last_type != NodeType.ACT:
            ctx.scene = node
            ctx.act = node
        else:
            ctx.scene = node
        ctx.index.append(node)
        return ctx

    @staticmethod
    def guess_formatting(text: str) -> InputType:
        return (
            InputType.HTML
            if text.startswith("<") and text.endswith(">")
            else InputType.MD
        )

    @functools.lru_cache(typed=True)
    def parse(
        self,
        text: str,
        title: str = None,
        *,
        input_type: InputType = None,
        tree: bool = True,
    ) -> Play:
        ctx = ParserContext()
        input_type = input_type or self.guess_formatting(text)
        text = html2text(text) if input_type is InputType.HTML else text
        for ix, line in enumerate(x for x in text.split("\n") if x.strip()):
            node = self.match(line, ix)
            append = self.check_direction(ctx, node)
            if append:
                self.check_linecount(ctx, node)
                handler = self.__parser_map[node.type]
                ctx = handler(ctx=ctx, node=node)

        return ctx.index.to_tree(title=title) if tree else ctx.index

    __call__ = parse


parser = Parser()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from html2text import html2text

from iambic import ast
from .html import render_html


def render_markdown(tree: ast.Play, with_toc: bool = False) -> str:
    """The renderer for a Markdown document.

    This is including for convenience. The guts of the document rendering is housed in
    :class:`~iambic.render.html.HTMLRenderer`

    Parameters
    ----------
    tree :
        The play tree, as loaded via :func:`~iambic.parser.parse` or :func:`~iambic.loader.load`
    with_toc : default False
        Optionally render a TOC for your tree.

    See Also
    --------
    :class:`iambic.render.html.HTMLRenderer`
    """
    return html2text(render_html(tree, with_toc=with_toc))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import inspect
import dataclasses
from typing import Type

import dataslots
import typic
import ujson as json


dataclass_sig = inspect.signature(dataclasses.dataclass)
typicklass_sig = inspect.signature(typic.klass)


def dataschema(_cls=None, **kwargs) -> Type:
    """A class-wrapper in the style of :py:func:`dataclasses.dataclass`.

    Other Parameters
    ----------------
    add_dict : True
        This is passed to :py:func:`dataslots.with_slots` and will be ignored if ``slots=False``
    add_weakref : bool, default False
        This is passed to :py:func:`dataslots.with_slots` and will be ignored if ``slots=False``
    frozen : bool, default False
        This is passed to :py:func:`dataclasses.dataclass`
    init : bool, default True
        This is passed to :py:func:`dataclasses.dataclass`
    repr : bool, default True
        This is passed to :py:func:`dataclasses.dataclass`
    eq : bool, default True
        This is passed to :py:func:`dataclasses.dataclass`
    order : bool, default False
        This is passed to :py:func:`dataclasses.dataclass`
    unsafe_hash : bool, default False
        This is passed to :py:func:`dataclasses.dataclass`
    """

    def wrap_cls(_cls):
        dcls = dataclasses.dataclass(
            _cls, **{x: y for x, y in kwargs.items() if x in dataclass_sig.parameters}
        )
        dcls.astuple = dataclasses.astuple
        dcls.replace = dataclasses.replace
        dcls.json = lambda cls: json.dumps(cls.asdict(), indent=2)
        dcls.json.__name__ = "json"
        dcls.json.__qualname__ = f"{dcls.__name__}.json"
        cls_ = typic.al(
            dataslots.with_slots(
                dcls,
                add_dict=kwargs.pop("add_dict", False),
                add_weakref=kwargs.pop("add_weakref", False),
            ),
            delay=kwargs.pop("delay", False),
        )
        cls_.__signature__ = inspect.signature(dcls)
        cls_.__annotations__ = dcls.__annotations__
        cls_.__qualname__ = dcls.__qualname__

        return cls_

    return wrap_cls if _cls is None else wrap_cls(_cls)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from collections.abc import Mapping
from types import MappingProxyType
from typing import Hashable, Iterator, Union, Tuple, List, Any, FrozenSet


class frozendict(Mapping):
    """An immutable, hashable mapping."""

    def __init__(
        self, obj: Union[Mapping, List[Tuple[str, Hashable]]] = None, **kwargs
    ):
        d = {x: self._freeze(y) for x, y in dict(obj or {}, **kwargs).items()}
        self.__r = f"{type(self).__name__}({repr(d)})"
        self.__d = MappingProxyType(d)
        self.__h = hash(frozenset(self.__d))

    @classmethod
    def _freeze(cls, o: Any) -> Hashable:
        if isinstance(o, Hashable):
            return o

        if isinstance(o, Mapping):
            return cls({x: cls._freeze(y) for x, y in o.items()})

        freezer = frozenset if isinstance(o, set) else tuple
        return freezer(cls._freeze(x) for x in o)

    def __repr__(self):
        return self.__r

    def __len__(self) -> int:
        return len(self.__d)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__d)

    def __getitem__(self, k: str) -> Union["frozendict", Hashable, Tuple, FrozenSet]:
        return self.__d.__getitem__(k)

    def __hash__(self):
        return self.__h



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import dataclasses
from typing import Tuple

import pytest

from iambic import ast, parse
from iambic.parse.text import ParserContext, _PreNode
from iambic.schema import frozendict


@pytest.mark.parametrize(
    argnames=("value", "expected"),
    argvalues=[
        ("# ACT I", ast.NodeType.ACT),
        ("### act I", ast.NodeType.ACT),
        ("### scene i", ast.NodeType.SCENE),
        ("## scene i", ast.NodeType.SCENE),
        ("# scene i", ast.NodeType.SCENE),
        ("# scene i. a foo in the forest.", ast.NodeType.SCENE),
        ("# epilogue", ast.NodeType.EPIL),
        ("## PROLOGUE", ast.NodeType.PROL),
        ("# Intermission", ast.NodeType.INTER),
        ("**FOO**", ast.NodeType.PERS),
        ("**Foo's Bar**", ast.NodeType.PERS),
        ("**Foobar 1**", ast.NodeType.PERS),
        ("**Another Foobar 1**", ast.NodeType.PERS),
        ("**Foo Bar Baz Buzz**", ast.NodeType.PERS),
        ("**Foo-Bar Hyph. the Third**", ast.NodeType.PERS),
        ("_Foo does a thing._", ast.NodeType.DIR),
        ("_Foo does", ast.NodeType.DIR),
        ("a thing._", ast.NodeType.DIR),
        ("_[Aside]_", ast.NodeType.ACTION),
        ("_[Stabs foo.]_", ast.NodeType.ACTION),
        ("_Enter Foo, bar._", ast.NodeType.ENTER),
        ("_Exit Foo, bar._", ast.NodeType.EXIT),
        ("_Exeunt severally._", ast.NodeType.EXIT),
        ("Literally anything else.", ast.NodeType.DIAL),
    ],
)
def test_text_match(value: str, expected: ast.NodeType):
    assert parse.text.match(value, 1).type == expected


@pytest.mark.parametrize(
    argnames=("ctx", "node", "expected_parent_text"),
    argvalues=[
        (ParserContext(), _PreNode(type=ast.NodeType.ACT, text="Act I"), "Act I"),
        (ParserContext(), _PreNode(type=ast.NodeType.SCENE, text="Scene I"), "Scene I"),
        (
            ParserContext(),
            _PreNode(type=ast.NodeType.EPIL, text="Epilogue"),
            "Epilogue",
        ),
        (
            ParserContext(),
            _PreNode(type=ast.NodeType.PROL, text="Prologue"),
            "Prologue",
        ),
    ],
)
def test_locale_handler(ctx: ParserContext, node: _PreNode, expected_parent_text: str):
    ctx = parse.text.locale_handler(ctx, node)
    assert ctx.parent.text == expected_parent_text
    ctx_node = getattr(ctx, node.type, ctx.act)
    assert ctx_node == ast.GenericNode(**dataclasses.asdict(node))


@pytest.mark.parametrize(
    argnames=("ctx", "node"),
    argvalues=[(ParserContext(), _PreNode(type=ast.NodeType.PERS, text="Foo"))],
)
def test_persona_handler(ctx: ParserContext, node: _PreNode):
    ctx = parse.text.persona_handler(ctx, node)
    assert ctx.character == node.to_generic()


generic_act = ast.GenericNode("act", "Act I", 0, 0)
generic_scene = ast.GenericNode("scene", "Scene II", 0, 0)


@pytest.mark.parametrize(
    argnames=("ctx", "node", "expected"),
    argvalues=[
        (
            ParserContext(generic_act, generic_act, generic_scene),
            _PreNode(ast.NodeType.DIAL, text=""),
            ast.GenericNode(
                "dialogue",
                text="",
                index=0,
                lineno=0,
                linepart=0,
                scene=generic_scene.resolve().id,
            ),
        ),
        (
            ParserContext(generic_act, generic_act, generic_scene),
            _PreNode(ast.NodeType.INTER),
            ast.GenericNode(
                "intermission",
                text=None,
                index=0,
                lineno=0,
                linepart=0,
                parent=generic_act.resolve().id,
            ),
        ),
    ],
)
def test_default_handler(ctx: ParserContext, node: _PreNode, expected):
    assert parse.text.default_handler(ctx, node).index.generic[-1] == expected


@pytest.mark.parametrize(
    argnames=("ctx", "node", "expected"),
    argvalues=[
        # a normal line.
        (ParserContext(), _PreNode(ast.NodeType.DIAL, text="Foo."), (1, 0, 1, 0)),
        # start of a line, up lineno & linepart.
        (ParserContext(), _PreNode(ast.NodeType.DIAL, text="Foo. ..."), (1, 1, 1, 1)),
        # within a line-continuation, up only linepart.
        (
            ParserContext(),
            _PreNode(ast.NodeType.DIAL, text="... Foo. ..."),
            (0, 1, 0, 1),
        ),
        # end of a line-continuation, up only linepart and reset linepart in context.
        (ParserContext(), _PreNode(ast.NodeType.DIAL, text="... Foo."), (0, 0, 0, 1)),
    ],
)
def test_check_linecount(
    ctx: ParserContext, node: _PreNode, expected: Tuple[int, int, int, int]
):
    parse.text.check_linecount(ctx, node)
    assert (ctx.lineno, ctx.linepart, node.lineno, node.linepart) == expected


def test_check_direction():
    ctx = ParserContext(index=ast.Index())
    ctx.index.append(
        ast.GenericNode(
            type="direction",
            text="_Foo",
            index=0,
            match=frozendict({"start": "_", "direction": "Foo"}),
        )
    )
    node = _PreNode(type=ast.NodeType.DIAL, text="foo")
    append = parse.text.check_direction(ctx, node)
    assert not append
    assert ctx.index.generic[-1].text == "_Foo foo"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from iambic import parse
from tests.static import MD_RAW


def test_data():
    parsed = parse.text(MD_RAW)
    data = parse.data(parsed.json())
    assert data == parsed



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from html2text import html2text
from iambic import parse, render
from tests.static import HTML, MD_RAW


def test_markdown():
    parsed = parse.text(MD_RAW)
    assert render.markdown(parsed, with_toc=True).strip() == html2text(HTML).strip()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import enum
from typing import List, Dict, Hashable, Tuple

import tablib

from iambic import ast


Row = List[Hashable]
Table = Dict[str, Row]
Matrix = List[Tuple[Hashable, ...]]


class Column(str, enum.Enum):
    """Predefined column names"""

    CHAR = "Dramatis Personae"
    APP = "First Appearance"
    CLINE = "Lines"
    PLINE = "Player Lines"
    PLAYR = "Player"
    SORT = "Sort"
    GSWAP = " [SW]"


class Marker(str, enum.Enum):
    """The character to use when marking a character as present in a scene."""

    SPEAK = "X"
    PRES = "O"
    NONE = ""


class Tabulator:
    """Generate a tabular representation of :py:class:`~iambic.ast.Play`

    Also known as a 'character map', this utility will generate a table
    which shows the number of lines for a given character and where they
    appear in the play.

    .. i.e.:
        | Dramatis Personae | Lines | I.i | I.ii | Player | Player Lines |
        ------------------------------------------------------------------
        | Johnny Appleseed  | 100   |  x  |   x  |  Jimmy |      75      |

    """

    @staticmethod
    def _tabulate_scene(
        scene: ast.NodeTree,
        node_column: List[str],
        cline_column: List[int],
        char_index: Dict[str, int],
        personae: Dict[str, ast.Persona],
    ):
        for child in scene.children:
            if isinstance(child, ast.Speech):
                persona = personae[child.persona]
                index = char_index[persona.name]
                node_column[index] = Marker.SPEAK.value
                cline_column[index] += child.num_lines
            elif isinstance(child, ast.Entrance):
                for pers in child.personae:
                    persona = personae[pers]
                    index = char_index[persona.name]
                    node_column[index] = Marker.PRES.value

    def tabulate(self, play: ast.Play) -> Table:
        """Generate a table for this play.

        The table is returned in the form of a Mapping.
        The keys are the header and the columns are the values.

        .. i.e.:
            {
                "foo": ['a', 'value'],
                "bar": ['another, 'value],
                ...
            }
        """
        table = dict()
        table[Column.CHAR.value] = list(x.name for x in play.personae)
        table[Column.APP.value] = list(x.index for x in play.personae)
        char_column = table[Column.CHAR.value]
        table[Column.CLINE.value] = list(0 for _ in char_column)
        cline_column = table[Column.CLINE.value]
        char_index = {y: x for x, y in enumerate(char_column)}
        personae = {x.id: x for x in play.personae}
        for act in play.children:
            # Epilogues and Prologues can be shaped like Scenes or Acts.
            # And can be top-level, like Acts.
            children = act.children
            if act.node.type in {
                ast.NodeType.EPIL,
                ast.NodeType.PROL,
            } and not isinstance(children[0], ast.NodeTree):
                children = [act]
            for scene in children:
                node = scene.node if isinstance(scene, ast.NodeTree) else scene
                table[node.col] = list(Marker.NONE.value for _ in char_column)
                if isinstance(scene, ast.NodeTree):
                    self._tabulate_scene(
                        scene=scene,
                        node_column=table[node.col],
                        cline_column=cline_column,
                        char_index=char_index,
                        personae=personae,
                    )

        return table

    @staticmethod
    def matrix(table: Table) -> Matrix:
        """Pivot a table into a 2-D array (matrix).

        The first entry is the header and the proceeding lines are values.

        .. i.e.:
            [
                ("foo", "bar"),
                ("a", "another"),
                ("value", "value"),
                ...
            ]
        """
        return [tuple(table.keys())] + list(zip(*table.values()))

    def dataset(self, table: Table) -> tablib.Dataset:
        """Transform a table into an instance of :py:class:`tablib.Dataset`."""
        pivot = self.matrix(table)
        headers = pivot.pop(0)
        return tablib.Dataset(*pivot, headers=headers)

    def __call__(self, play: ast.Play) -> tablib.Dataset:
        table = self.tabulate(play)
        return self.dataset(table)


tabulate = Tabulator()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import dataclasses
import functools
import inspect
from typing import ClassVar, Optional, Union, Tuple, Any, Mapping, Type, Match, List

import inflection
import typic

from iambic import schema, roman
from iambic.schema import frozendict
from .base import NodeType, NodeMixin, DEFINITIONS


__all__ = (
    "Act",
    "Scene",
    "Prologue",
    "Epilogue",
    "Intermission",
    "Persona",
    "Entrance",
    "Exit",
    "Action",
    "Direction",
    "Dialogue",
    "Speech",
    "ResolvedNode",
    "ChildNode",
    "NodeTree",
    "Play",
    "MetaData",
    "GenericNode",
    "node_coercer",
    "isnodetype",
)


@schema.dataschema(frozen=True)
class Act(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Act"]
    )
    type: ClassVar[NodeType] = NodeType.ACT
    index: int
    text: str
    num: int

    @property
    @functools.lru_cache(1)
    def id(self):
        return inflection.parameterize(f"act-{self.col}")

    @property
    @functools.lru_cache(1)
    def col(self):
        return roman.numeral(self.num)

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Act":
        numeral = node.pieces[1]
        num = int(numeral) if numeral.isdigit() else roman.integer(numeral)
        return cls(index=node.index, text=node.match_text, num=num)


@schema.dataschema(frozen=True)
class Scene(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Scene"]
    )
    type: ClassVar[NodeType] = NodeType.SCENE
    index: int
    text: str
    num: int
    act: str
    setting: Optional[str]

    @property
    @functools.lru_cache(1)
    def id(self) -> str:
        return inflection.parameterize(f"{self.act}-scene-{roman.numeral(self.num)}")

    @property
    @functools.lru_cache(1)
    def col(self) -> str:
        if self.act in {NodeType.PROL, NodeType.EPIL}:
            pre = self.act[0].upper()
        else:
            pre = self.act.split("-")[-1].upper()
        return f"{pre}.{roman.numeral(self.num).lower()}"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Scene":
        numeral = node.pieces[1]
        num = int(numeral) if numeral.isdigit() else roman.integer(numeral)
        setting = " ".join(node.pieces[2:]) if len(node.pieces) > 2 else None
        parent = node.parent or node.act or node.scene
        parent = parent.resolve() if hasattr(parent, "resolve") else parent
        return cls(
            index=node.index, text=node.match_text, num=num, act=parent, setting=setting
        )


@schema.dataschema(frozen=True)
class Prologue(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Prologue"]
    )
    type: ClassVar[NodeType] = NodeType.PROL
    index: int
    text: str
    setting: Optional[str]
    act: str = None

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"{self.act}-prologue" if self.act else "prologue"

    @property
    @functools.lru_cache(1)
    def col(self):
        pre = f"{self.act.split('-')[-1].upper()}." if self.act else ""
        return f"{pre}P"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Prologue":
        setting = " ".join(node.pieces[1:]) if len(node.pieces) > 1 else None
        return cls(
            index=node.index, text=node.match_text, setting=setting, act=node.parent
        )


class Epilogue(Prologue):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Epilogue"]
    )
    type: ClassVar[NodeType] = NodeType.EPIL

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"{self.act}-epilogue" if self.act else "epilogue"

    @property
    @functools.lru_cache(1)
    def col(self):
        pre = f"{self.act.split('-')[-1].upper()}." if self.act else ""
        return f"{pre}E"


@schema.dataschema(frozen=True)
class Intermission(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Intermission"]
    )
    type: ClassVar[NodeType] = NodeType.INTER
    index: int
    text: str
    act: str

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"intermission"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Intermission":
        return cls(index=node.index, text=node.match_text, act=node.parent)

    @property
    @functools.lru_cache(1)
    def col(self):
        return "INT"


@schema.dataschema(frozen=True)
class Persona(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Persona"]
    )
    type: ClassVar[NodeType] = NodeType.PERS
    index: int
    text: str
    name: str
    short: str = None

    def __eq__(self, other) -> bool:
        return other.name == self.name if hasattr(other, "name") else False

    @property
    @functools.lru_cache(1)
    def id(self):
        return inflection.parameterize(self.name)

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Persona":
        return cls(
            index=node.index,
            text=node.match_text,
            name=inflection.titleize(node.match_text),
        )


@schema.dataschema(frozen=True)
class Entrance(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Entrance"]
    )
    type: ClassVar[NodeType] = NodeType.ENTER
    index: int
    text: str
    scene: str
    personae: Tuple[str] = dataclasses.field(default_factory=tuple)

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"{self.scene}-entrance-{self.index}"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Entrance":
        return cls(index=node.index, text=node.match_text, scene=node.parent)


class Exit(Entrance):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Exit"]
    )
    type: ClassVar[NodeType] = NodeType.EXIT

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"{self.scene}-exit-{self.index}"


@schema.dataschema(frozen=True)
class Action(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Action"]
    )
    type: ClassVar[NodeType] = NodeType.ACTION
    action: str
    persona: str
    scene: str
    index: int

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"{self.scene}-{self.persona}-action-{self.index}"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Action":
        return cls(
            action=node.match_text,
            persona=node.parent,
            scene=node.scene,
            index=node.index,
        )


@schema.dataschema(frozen=True)
class Direction(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Direction"]
    )
    type: ClassVar[NodeType] = NodeType.DIR
    action: str
    scene: str
    index: int
    stop: bool = True

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"{self.scene}-direction-{self.index}"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Direction":
        return cls(action=node.match_text, scene=node.parent, index=node.index)


@schema.dataschema(frozen=True)
class Dialogue(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Dialogue"]
    )
    type: ClassVar[NodeType] = NodeType.DIAL
    line: str
    persona: str
    scene: str
    index: int
    lineno: int
    linepart: int = 0

    @property
    @functools.lru_cache(1)
    def id(self):
        return f"{self.persona}-dialogue-{self.lineno}"

    @classmethod
    def from_node(cls, node: "GenericNode") -> "Dialogue":
        return cls(
            node.match_text,
            node.parent,
            node.scene,
            index=node.index,
            lineno=node.lineno,
        )


@schema.dataschema(frozen=True)
class Speech(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Speech"]
    )
    type: ClassVar[NodeType] = NodeType.SPCH
    persona: str
    scene: str
    speech: Tuple[Union[Dialogue, Action, Direction]]
    index: int

    @property
    @functools.lru_cache(1)
    def linerange(self) -> Tuple[int, int]:
        linenos = sorted([x.lineno for x in self.speech if hasattr(x, "lineno")])
        return linenos[0], linenos[-1]

    @property
    @functools.lru_cache(1)
    def num_lines(self) -> int:
        x, y = self.linerange
        # line count starts at 1
        # a range of 1 - 1 is 1 line
        return y - (x - 1)

    @property
    @functools.lru_cache(1)
    def id(self) -> str:
        return (
            f"{self.scene}-{self.persona}-speech-{'-'.join(map(str, self.linerange))}"
        )


ResolvedNode = Union[
    Act,
    Scene,
    Prologue,
    Epilogue,
    Persona,
    Entrance,
    Exit,
    Action,
    Direction,
    Dialogue,
    Speech,
    Intermission,
]

ChildNode = ResolvedNode


@schema.dataschema(frozen=True, delay=True)
class NodeTree(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["NodeTree"]
    )
    type: ClassVar[NodeType] = NodeType.TREE
    node: ResolvedNode
    children: Tuple[ChildNode] = dataclasses.field(default_factory=tuple)
    personae: Tuple[str] = dataclasses.field(default_factory=tuple)

    @property
    @functools.lru_cache(1)
    def cols(self) -> Tuple[Any]:
        return tuple(
            getattr(x, "node", x).col
            for x in self.children
            if isinstance(x, (NodeTree, Prologue))
        )

    @property
    @functools.lru_cache(1)
    def id(self) -> str:
        return f"{self.node.id}-tree"


ChildNode = Union[
    Act,
    Scene,
    Prologue,
    Epilogue,
    Persona,
    Entrance,
    Exit,
    Action,
    Direction,
    Dialogue,
    Speech,
    Intermission,
    NodeTree,
]


@schema.dataschema(frozen=True)
class MetaData(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["MetaData"]
    )
    type: ClassVar[NodeType] = NodeType.META
    rights: str = "Creative Commons Non-Commercial Share Alike 3.0"
    language: str = "en-GB-emodeng"
    publisher: str = "Published w/ ❤️ using iambic - https://pypi.org/project/iambic"
    title: str = None
    subtitle: str = None
    edition: int = 1
    author: str = "William Shakespeare"
    editors: Tuple[str] = dataclasses.field(default_factory=tuple)
    tags: Tuple[str] = dataclasses.field(default_factory=tuple)

    @functools.lru_cache(1)
    def asmeta(self):
        dikt = {
            "creator": [{"type": "author", "text": self.author}],
            "contributor": [{"type": "editor", "text": "MIT"}],
            "rights": self.rights,
            "language": self.language,
            "publisher": self.publisher,
            "subject": ["Shakespeare"],
        }
        if self.title:
            dikt["title"] = [
                {"type": "main", "text": self.title},
                {"type": "edition", "text": self.edition},
            ]
            if self.subtitle:
                dikt["title"].append({"type": "subtitle", "text": self.subtitle})
        dikt["contributor"].extend(
            ({"type": "editor", "text": x} for x in self.editors or [])
        )
        dikt["subject"].extend(self.tags or [])

        return dikt


@schema.dataschema(frozen=True, delay=True)
class Play(NodeMixin):
    __static_definition__: ClassVar[schema.frozendict] = schema.frozendict(
        DEFINITIONS["Play"]
    )
    type: ClassVar[NodeType] = NodeType.PLAY
    children: Tuple[NodeTree] = dataclasses.field(default_factory=tuple)
    personae: Tuple[Persona] = dataclasses.field(default_factory=tuple)
    meta: MetaData = dataclasses.field(default_factory=MetaData)

    @property
    @functools.lru_cache(1)
    def id(self) -> str:
        return inflection.parameterize(f"{self.meta.title}-play")


@dataclasses.dataclass(frozen=True)
class GenericNode(NodeMixin):
    """The root-object of a script.

    A script ``Node`` represents a single line of text in a script.
    """

    __resolver_map__: ClassVar[Mapping[NodeType, ResolvedNode]] = frozendict(
        {
            NodeType.ACT: Act,
            NodeType.ACTION: Action,
            NodeType.DIAL: Dialogue,
            NodeType.DIR: Direction,
            NodeType.ENTER: Entrance,
            NodeType.EPIL: Epilogue,
            NodeType.EXIT: Exit,
            NodeType.INTER: Intermission,
            NodeType.META: MetaData,
            NodeType.PERS: Persona,
            NodeType.PLAY: Play,
            NodeType.PROL: Prologue,
            NodeType.SCENE: Scene,
            NodeType.SPCH: Speech,
            NodeType.TREE: NodeTree,
        }
    )
    # Minimum data for a Node
    type: NodeType
    text: str
    index: int
    # Additional fields which may be present
    lineno: int = None
    linepart: int = None
    # Given by text parser
    # If reading from JSON, we don't have/need this,
    # it will be provided inherently by the data-structure
    # on resolution-time.
    match: frozendict = dataclasses.field(default_factory=frozendict)
    parent: str = None
    act: str = None
    scene: str = None

    @classmethod
    @functools.lru_cache(maxsize=None)
    def resolvable(cls, obj: Type) -> bool:
        values = cls.__resolver_map__.values()
        return obj in values

    @classmethod
    @functools.lru_cache(maxsize=1)
    def sig(cls) -> inspect.Signature:
        return inspect.signature(cls)

    @functools.lru_cache(maxsize=None)
    def resolve(self) -> ResolvedNode:
        """Resolve a GenericNode into a typed, "resolved" Node.

        Raises
        ------
        TypeError
            If the NodeType provided has no resolved Node mapping.
        """
        try:
            return self.__resolver_map__[self.type].from_node(self)
        except KeyError:
            raise TypeError(
                f"Unrecognized node-type <{self.type}> for text <{self.text}>. "
                f"Valid types are: {tuple(self.__resolver_map__)}"
            )

    @property
    @functools.lru_cache(maxsize=1)
    def pieces(self) -> List[str]:
        return self.match_text.split()

    @property
    @functools.lru_cache(maxsize=1)
    def match_text(self) -> str:
        return self.match[self.type] if self.match else self.text


_RESOLVABLE = set(GenericNode.__resolver_map__.values())
_candidates = _RESOLVABLE.union(ChildNode.__args__)
_candidates.add(NodeTree)


def node_coercer(value: Any) -> Optional[ResolvedNode]:
    if type(value) in _RESOLVABLE or value is None:
        return value
    if isinstance(value, GenericNode):
        return value.resolve()

    if not isinstance(value, Mapping):
        value = typic.coerce(value, dict)

    handler: ResolvedNode = GenericNode.__resolver_map__[value.pop("type")]
    return handler(**value)


@functools.lru_cache(maxsize=None)
def isnodetype(obj: Type) -> bool:
    is_valid = (
        obj is ResolvedNode
        or obj is GenericNode
        or obj in _candidates
        or (
            getattr(obj, "__origin__", None) is Union
            and set(getattr(obj, "__args__", set())).issubset(_candidates)
        )
    )
    return is_valid



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import functools
from collections import deque, defaultdict
from typing import (
    Deque,
    Sequence,
    DefaultDict,
    Dict,
    Type,
    Mapping,
    Tuple,
    Optional,
    List,
    Set,
    Union,
)


import iambic.ast.node as ast

__all__ = ("Index",)


class Index(Deque[ast.ResolvedNode]):
    """An ordered collection of nodes in a script.

    ``Parser.parse`` builds a list of :class:`GenericNode`, which can be resolved into the specific
    data-type for the :class:`NodeType` assigned.
    """

    def __init__(self, initlist: Sequence[ast.GenericNode] = None):
        initlist = initlist or []
        super().__init__(initlist)
        self.generic = deque()
        self.__type_map: DefaultDict[
            ast.NodeType, Dict[str, ast.ResolvedNode]
        ] = defaultdict(dict)

    def __setitem__(self, key, value: ast.GenericNode):
        if isinstance(value, ast.GenericNode):
            self.generic[key] = value
            value = value.resolve()
        self.__type_map[value.type][value.id] = value
        super().__setitem__(key, value)

    def __getitem__(
        self, item: Union[ast.NodeType, ast.ResolvedNode, ast.GenericNode, int, str]
    ):
        if isinstance(item, int):
            return super().__getitem__(item)
        if isinstance(item, ast.NodeType):
            return self.__type_map[item]
        if isinstance(item, ast.GenericNode):
            item = item.resolve()
        if ast.GenericNode.resolvable(type(item)):
            return self.__type_map[item.type].get(item.id)

    def append(self, item: ast.GenericNode) -> None:
        if isinstance(item, ast.GenericNode):
            self.generic.append(item)
            item = item.resolve()
        self.__type_map[item.type][item.id] = item
        super().append(item)

    def appendleft(self, item: ast.GenericNode) -> None:
        if isinstance(item, ast.GenericNode):
            self.generic.appendleft(item)
            item = item.resolve()
        self.__type_map[item.type] = dict(
            **{item.id: item}, **self.__type_map[item.type]
        )
        super().appendleft(item)

    def get(self, node_type: Type[ast.ResolvedNode]) -> Mapping[str, ast.ResolvedNode]:
        return self.__type_map[node_type.type]

    def personae(self) -> Tuple[ast.Persona, ...]:
        return tuple(self.get(ast.Persona).values())

    def acts(self) -> Tuple[ast.Act, ...]:
        return tuple(self.get(ast.Act).values())

    def scenes(self) -> Tuple[ast.Scene, ...]:
        return tuple(self.get(ast.Scene).values())

    def prologues(self) -> Tuple[ast.Prologue, ...]:
        return tuple(self.get(ast.Prologue).values())

    def epilogues(self) -> Tuple[ast.Epilogue, ...]:
        return tuple(self.get(ast.Epilogue).values())

    def dialogue(self) -> Tuple[ast.Dialogue, ...]:
        return tuple(self.get(ast.Dialogue).values())

    def intermission(self) -> Optional[ast.Intermission]:
        inter = tuple(self.get(ast.Intermission).values())
        return inter[-1] if inter else None

    def get_speeches(self) -> List[ast.Speech]:
        persona, scene, speech, speeches = None, None, [], []
        for node in self:
            if type(node) in {ast.Scene, ast.Prologue, ast.Epilogue, ast.Persona}:
                if isinstance(node, ast.Persona):
                    if persona and scene and speech:
                        speeches.append(
                            ast.Speech(
                                persona.id, scene.id, tuple(speech), speech[0].index
                            )
                        )
                    persona, speech = node, []
                elif type(node) in {ast.Scene, ast.Prologue, ast.Epilogue}:
                    if persona and scene and speech:
                        speeches.append(
                            ast.Speech(
                                persona.id, scene.id, tuple(speech), speech[0].index
                            )
                        )
                    persona, scene, speech = None, node, []
            elif type(node) in {ast.Dialogue, ast.Action, ast.Direction}:
                speech.append(node)

        if persona and scene and speech:
            speeches.append(
                ast.Speech(persona.id, scene.id, tuple(speech), speech[0].index)
            )
        return speeches

    def filter_directions(self, speeches: Set[ast.Speech]) -> Set[ast.Direction]:
        speech_directions = set(
            y for x in speeches for y in x.speech if isinstance(y, ast.Direction)
        )
        return set(self.get(ast.Direction).values()) - speech_directions

    def get_scene_trees(self, sort: bool = True) -> List[ast.NodeTree]:
        speeches = set(self.get_speeches())
        directions = self.filter_directions(speeches)
        entrances = set(self.get(ast.Entrance).values())
        exits = set(self.get(ast.Exit).values())
        children = speeches | directions | entrances | exits
        scenes = []
        for node in set(self.scenes()) | set(self.prologues()) | set(self.epilogues()):
            child_nodes = set(x for x in children if x.scene == node.id)
            personae = set(
                x.persona for x in child_nodes if x.type == ast.NodeType.SPCH
            )
            if child_nodes:
                tree = ast.NodeTree(
                    node=node,
                    children=tuple(sorted(child_nodes, key=self.node_sort)),
                    personae=tuple(personae),
                )
                scenes.append(tree)
                children -= child_nodes
        if sort:
            scenes.sort(key=self.node_sort)
        return scenes

    def get_act_trees(self, scenes: List[ast.NodeTree]) -> List[ast.NodeTree]:
        scenes = set(scenes)
        acts = {x for x in scenes if x.node.act is None}
        scenes -= set(acts)
        intermission = self.intermission()
        prols = tuple(x for x in self.prologues() if x.act is None)
        epils = tuple(x for x in self.epilogues() if x.act is None)
        for act in set(prols + self.acts() + epils):
            children = set(x for x in scenes if x.node.act == act.id)
            if children:
                scenes -= children
                # Add the intermission if there is one.
                if intermission and intermission.act == act.id:
                    children.add(intermission)
                nodes = tuple(sorted(children, key=self.node_sort))
                acts.add(ast.NodeTree(act, nodes))
        acts = list(sorted(acts, key=self.node_sort))
        return acts

    def to_tree(self, title: str = None) -> ast.Play:
        # Build the Act-level trees
        scenes = self.get_scene_trees()
        # Build the Play-level trees
        acts = self.get_act_trees(scenes)
        # Put it all together
        play = ast.Play(tuple(acts), tuple(self.personae()), ast.MetaData(title=title))
        return play

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def node_sort(node):
        return getattr(node, "node", node).index



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from iambic import parse, render
from tests.static import HTML, MD_RAW


def test_html():
    parsed = parse.text(MD_RAW)
    rendered = render.html(parsed)
    assert rendered.replace(" ", "").split() == HTML.replace(" ", "").split()
